#ifndef mat_h__
#define mat_h__

#define inline
#include "types.h"
#include<stdio.h>
#include<assert.h>
template <typename T>
struct Mat
{
  Mat() : m_mat( NULL )
  {
    for( int i = 0; 2 != i; ++i )
      m_size[i] = 0;
  }

  Mat( uint row, uint col, float const *yo) : m_mat( NULL )
  {
    setSize( row, col );
    uint togo( row * col );
    for( int count = 0; togo != count; ++count )
      m_mat[count] = yo[count];
  }
  
  Mat( Mat const &a ) : m_mat( 0 ) // we need that, set size checks for it before deleting
  {
    size( a.size() );
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      m_mat[c] = a.m_mat[c];
  }
  ~Mat()
  {
    if( m_mat )
      delete [] m_mat;
  }
  /**
   * init gradualy a vector (say, [3, 1] will be .25, .5, .75)
   */
  void initStep();

  Mat &operator=( Mat const &a )
  {
    size( a.size() );
    for( int c = 0, togo = a.m_size[0] * a.m_size[1]; togo != c; ++c )
      m_mat[c] = a.m_mat[c];

  }

  uint m_size[2];

  Mat<T> &f( T f( T const & , T const & ), Mat<T> const &a );
  Mat<T> &operator-=( Mat<T> const & a ) ;
  Mat<T> const operator-( Mat<T> const & a ) const;


  /**
   * transpose
   */
  Mat<T> const t() const;

  void rand( T const & low, T const & hi );
  
  inline Mat<T> f( T f( T const & ) )
  {
    Mat<T> rv( *this );
    
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      rv.m_mat[c] = f( rv.m_mat[c] );

    return rv;
  }
  inline Mat<T> f( T f( T ) )
  {
    Mat<T> rv( *this );
    
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      rv.m_mat[c] = f( rv.m_mat[c] );

    return rv;
  }
  static inline Mat<T> Fixed( uint r, uint c, T f() );
  static inline Mat<T> Ones( uint const *size );
  static inline Mat<T> Zeros( uint r, uint c );
  static inline Mat<T> I( uint r, uint c );


  uint const inline *size() const { return getSize(); };

  inline void size( uint n ) { size( n, 1 ); };
  uint const inline *getSize() const { return m_size; }
  
  Mat<T> const &operator*=( T const & d );

  Mat<T> const operator*( T const & d ) const
  {
    Mat<T> rv( *this );

    rv *= d;
    return rv;
  }

  inline void size( uint const *size )
  {
    setSize( size[0], size[1] );
  }
  inline void size( uint r, uint c )
  {
    setSize( r, c );
  }
  inline void setSize( uint i, uint j )
  {
    //if( ( m_size[0] == i ) 
    //	&& (m_size[1] == j) )
    //  return;

    m_size[0] = i;
    m_size[1] = j;
    if( m_mat )
      delete [] m_mat;

    m_mat = new T[ i * j ];
  };

  T & operator()( uint i, uint j ) const
  {
    return m_mat[ i * m_size[1] + j];
  }


  
  struct Row
  {
 
    T &operator[]( uint i )
    {
      assert( i < size );
     
      return m_[i];
    }
    T operator[]( uint i ) const
    {
     assert( i < size );
     return m_[i];
    }
    T *m_;
    //for debug
    uint size;
    Row( T *p, uint size ) : m_( p ), size(size) {};
  };

  Row operator[]( uint i )
  {
    assert( i < m_size[0] );
    return Row( m_mat + i * m_size[1], m_size[1] );
  }
  Row const operator[]( uint i ) const
  {
    assert( i < m_size[0] );
    return Row( m_mat + i * m_size[1], m_size[1] );

  }

  Mat<T> const operator*( Mat<T> const & a ) const
  {
    Mat<T> rv = Zeros( size()[0], a.size()[1] );

    assert( size()[1] == a.size()[0] );
    for( uint colRes = 0; rv.size()[1] != colRes; ++colRes )
      for( uint rowRes = 0; rv.size()[0] != rowRes; ++rowRes )
	for( uint intCount = 0; a.size()[0] != intCount; ++intCount )
	{
	  float const &t1( (*this)[rowRes][intCount] );
	  Row const &arow( a[intCount] );
	  float const &t2( arow[colRes] );

	  float &store( rv[rowRes][colRes] );
	  
	  store += t1 * t2;
	}
    
    return rv;
  }

  //  private:
  T * m_mat;
};

#include <mat.inl>

#endif
