#include <assert.h>
#include "mat.h"
#include "NN.h"
#include <iostream>


typedef Mat<float> Matf;

/*
** activation function prototype and its derivative 
*/ 

float theta( float );
float theta_( float );

template <typename T>
NeuralNet<T>::NeuralNet( uint nLayer, uint const * layerSizes )
{
  assert( nLayer >= 2 ); 
  m_nLayers = nLayer;
  m_layersSizes = new uint[nLayer];
  m_vs = new Mat<T>[nLayer];
  m_ys = new Mat<T>[nLayer];

  m_weights = new Mat<T>[nLayer - 1];

  for( int layerI = 0; nLayer != layerI; ++layerI )
  {
    m_layersSizes[layerI] = layerSizes[layerI];

    m_vs[layerI].size( layerSizes[layerI] );
    m_ys[layerI].size( layerSizes[layerI] );

    if( layerI >= 1 )
      m_weights[layerI - 1].size( layerSizes[layerI], layerSizes[layerI - 1] );
      
  }
}

template <typename T>
NeuralNet<T>::~NeuralNet()
{

  delete [] m_ys;
  delete [] m_vs;

  delete [] m_weights;
  delete [] m_layersSizes;
}

template <typename T> 
std::basic_ostream<char> &
operator<<( std::basic_ostream<char> &o, Mat<T> const & m )
{
  for( int i = 0; m.m_size[0] != i; ++i )
  {
    o << ( i ? "   " : "[  ");

    for( int j = 0; m.m_size[1] != j; ++j )
    {
      o << m(i, j);
      if( ( m.m_size[1] - 1 ) != j )
	o << ", ";
    }
	
    o << ( ( ( m.m_size[0] - 1 ) == i ) ? "  ]" : ";" );
    o << std::endl;
  }
}

template <typename T>
void NeuralNet<T>::instruct( T const * in, T const * expected )
{

  present( in );

  static float const thau( 0.04 );
  /*
  ** back propagation
  */

  Mat<T> *ds = new Mat<T>[m_nLayers - 1];

#if 1
  // compute exit error
  ds[m_nLayers - 2] = Mat<T>( m_ys[m_nLayers - 1].size()[0], 1, expected ) - m_ys[m_nLayers - 1];

  // propagate
  for( int layer_i = m_nLayers - 3; layer_i >= 0; --layer_i )
  {
    Mat<T> const &d_next( ds[layer_i + 1] );
    Mat<T> const &weights( m_weights[layer_i + 1] );

    Mat<T> const t_weights( weights.t() );
    Mat<T> const newDs( t_weights * d_next );

    ds[layer_i] = newDs;
  }
#else
  Mat<T> expectedMat = Mat<T>( m_ys[m_nLayers - 1].size()[0], 1, expected );

  for( int weightLayer = m_nLayers - 2; weightLayer >= 0; --weightLayer )
  {
    Mat<T> const weights( m_nLayers - 2 != weightLayer ? m_weights[weightLayer + 1] : Mat<T>::I( m_ys[m_nLayers - 1].size()[0], m_ys[m_nLayers - 1].size()[0]  ) );
    Mat<T> const e( expectedMat - m_ys[weightLayer == m_nLayers - 2 ? m_nLayers - 1 : weightLayer + 2] );
    Mat<T> const t_weights( weights.t() );
    Mat<T> const newDs( t_weights * e );
    // std::cout << "e" << e;
    // std::cout << "w" << weights;
    // std::cout << "ds" << newDs;

    ds[weightLayer + 1] = newDs;
    expectedMat = ds[weightLayer + 1];    
  }

#endif
  // dump();
  // std::cout << "expected" << Mat<T>( m_ys[m_nLayers - 1].size()[0], 1, expected );
  // for( int i = 0; m_nLayers - 1 != i; ++i )
  //{
  //  std::cout << "ds[ " << i << " ]"<< std::endl;
  //  std::cout << ds[i];
  //  
 // }
  int I1 = 9;
  for( int weightLayer = 0; m_nLayers - 1 != weightLayer; ++weightLayer )
  {
    Mat<T> &weights( m_weights[weightLayer] );

    for( int j = 0; weights.size()[0] != j; ++j )
    {
      float const &d( ds[weightLayer][j][0] );

      
	
      for( int i = 0; weights.size()[1] != i; ++i )
      {
	float const dfde( theta_( m_vs[weightLayer + 1][j][0] ) );
	float const &yj( m_ys[weightLayer + 1][j][0] );
	float &w( weights[j][i] );
	//	w += thau * d * dfde * m_ys[weightLayer + 1][j][0];
	w += thau * d * dfde * m_ys[weightLayer][i][0];
      }
    }
    //dump();
    int I = 0;
  }

  delete [] ds;
  
#if 0
  Mat<T> prevD;
  for( int weightLayer = m_nLayers - 2; weightLayer >= 0; --weightLayer )
  {
    Mat<T> &wn = m_weights[weightLayer];
    Mat<T> nextD = Mat<T>::Zeros( wn.size()[1], 1 );

    for( int i = 0; i != wn.size()[1]; ++i ) // from
      for( int j = 0; j != wn.size()[0]; ++j ) // to
      {
	float &wjin( wn[j][i] );
	float d;

	Mat<T> const &v( m_vs[weightLayer + 1] );
	float const &vjn( v[j][0] );

	if( ( m_nLayers - 2 ) == weightLayer )
	{
	  assert( prevD.size()[0] == 0 ); 
	  assert( prevD.size()[1] == 0 );
	  float const ejn( in[j] - m_ys[weightLayer + 1][j][0] );
	  d = - ejn * theta_( vjn );
	} else
	{

	  //float sumEwkjn( 0.0f );
	  //for( int k = 0; k != wn.size()[0]; ++k )
	  assert( prevD.size()[0] == wn.size()[0] );
	  d = - theta_( vjn ) * prevD[j][0];
	}
	nextD[i][0] += d * wjin;
	float const deltaWjin( - thau * d * m_ys[weightLayer /** tocheck (2) */ ][i][0] );
	wjin += deltaWjin;
      }

    prevD = nextD;
  }
#endif
#if 0

  float precSumD_w[1024];
  uint numberOfK( 0 );

  for( int weightLayer = m_nLayers; weightLayer >= 0; --weightLayer )
  {
    Mat<float> & weights( m_weights[weightLayer] );
    bool const isOutputLayer( weightLayer == m_nInnerLayers );

    uint const nj( isOutputLayer
		   ? m_nOut 
		   : m_innerLayersSizes[weightLayer] );
    uint const ni( weightLayer == 0
		   ? m_nIn
		   : m_innerLayersSizes[weightLayer - 1]);

    float * const y( isOutputLayer ? m_out : m_ys[weightLayer] );
    float * const v( m_vs[weightLayer] );


    for( int j = 0; nj != j; ++j )
    {
      float new_precSumD_wj( 0.0f );
      for( int i = 0; ni != i; ++i )
      {
	float const d( theta_(v[j])
		       * ( isOutputLayer 
			   ? expected[j] - m_out[j]
			   : precSumD_w[j] ) );
	new_precSumD_wj += d * m_weights[weightLayer](j, i);
	
	float dwji = thau * d * y[i];
      }
      precSumD_w[j] = new_precSumD_wj;
    }

    numberOfK = ni;
  }

#endif
}

template <typename T>
void NeuralNet<T>::present( T const * in )
{
  m_ys[0] = Mat<T>( m_ys[0].size()[0], 1, in );


  for( int weightLayer = 0; weightLayer != ( m_nLayers - 1 ); ++weightLayer )
  {
    Mat<T> const &prevExit( m_ys[weightLayer] );

    Mat<T> &nextEntries( m_vs[weightLayer + 1] );
    nextEntries = m_weights[weightLayer] * prevExit;

    {
      Mat<T> &nextExit( m_ys[weightLayer + 1] );
      nextExit = nextEntries.f( theta );
      int I = 0;
    }
  }
}

template <>
void NeuralNet<float>::initRandom( float const & low, float const & hi )
{
  for( int weightLayer = 0; weightLayer < ( m_nLayers - 1 ); ++weightLayer )
    m_weights[weightLayer].rand( low, hi );
    
}

template <typename T>
void NeuralNet<T>::initFlat( T const & i )
{
  for( int weightLayer = 0; weightLayer < ( m_nLayers - 1 ); ++weightLayer )
    m_weights[weightLayer] = Mat<T>::Ones( m_weights[weightLayer].size() ) * i;
}

template <typename T>
void NeuralNet<T>::initStep()
{
  for( int weightLayer = 0; weightLayer != ( m_nLayers - 1 ); ++weightLayer )
    m_weights[weightLayer].initStep();
}

template class NeuralNet<float>;
