#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

//#include <vector>
#include<iostream>
#include<math.h>

/*
** so I think in the general case the best way of handling that it to offer the possibility to
** to solve by groups of points. 
*
** I propose a solution for an arbitrary number of points, which takes a number of arbitrary point for each 
** path snippet.
*
** I will then blend those solution together.
*
** the solution require all the n-uple consecutive point to lay on the
** axis formed by the extrimities of the snippet. 
*
** What this means is that if you are to use a quadric, all three
** consecutive point must form an acute triangle.
*
** I think we could solve that by adding a constraint points where this is not the case. 
*
** germain
*/

#include <assert.h>
#include <math.h>

typedef unsigned int uint;
#define NULL 0
#define sqr( x ) ( ( x ) * ( x ) )

template <typename T>
class Vector
{
  T m_pos[3];
public:
  Vector(){};

  Vector( T const &x, T const &y, T const &z )
  {
    m_pos[0] = x;
    m_pos[1] = y;
    m_pos[2] = z;
  }

  /*\
   * @brief dot product
   */

  inline T const dot( Vector const &a ) const
  {
    T returnValue( .0f );
 
    for( int i = 0; 3 != i; ++i )
      returnValue += (*this)[i] * a[i];
    
    return returnValue;
  }

  inline T operator*( Vector const &a ) const
  {
    return this->dot( a );
  }

  /*\
   * @brief cross product
   */
  inline Vector cross( Vector const &v ) const
  {
    return Vector( (*this)[1] * v[2] - (*this)[2] * v[1], 
		   (*this)[2] * v[0] - (*this)[0] * v[2],
		   (*this)[0] * v[1] - (*this)[1] * v[0] );
  }
  inline Vector operator^( Vector const & v ) const
  {
    return this->cross( v );
  }
  inline Vector &operator+=( Vector const & v )
  {
    for( int i = 0; 3 != i; ++i )
      m_pos[i] += v[i];
    return *this;
  }

  inline Vector operator+( Vector const & v ) const
  {
    Vector rv( *this );
    return rv += v;
  }

  inline Vector &operator*=( T const & a )
  {
    for( int i = 0; 3 != i; ++i)
      m_pos[i] *= a;
    return *this;
  }

  inline Vector &operator-=( Vector const &a )
  {
    return (*this) += a * -1.0f;
  }

  inline Vector operator-( Vector const &a ) const
  {
    return *this + ( a * -1.0f );
  }

  inline Vector operator*( T const & a ) const
  {
    Vector rv( *this );
    return rv *= a;
  }

  inline Vector operator/( T const & a ) const
  {
    return *this * ( 1.0f / a );
  }
  
  inline T dist( Vector const &a ) const
  {
    return sqrt( dist2( a ) );
  }

  inline T length() const
  {
    return dist( Vector( 0, 0, 0) );
  }

  inline Vector unit() const
  {
    return *this / length();
  }

  inline T dist2( Vector const &a ) const
  {
    T rv( .0f );

    for( int i = 0; 3 != i; ++i )
      rv += sqr( m_pos[i] - a.m_pos[i] );

    return rv;
  }

  inline T const &operator[]( int indice ) const { return m_pos[indice]; };
  inline T &operator[]( int indice ) { return m_pos[indice]; };
};

template <typename T>
struct Mat
{

  Mat() : m_mat( NULL )
  {
    for( int i = 0; 2 != i; ++i )
      m_size[i] = 0;
  }

#if 0
  template <int vdim>
  Mat( Vertex<vdim,T> const &v ) : m_mat( NULL )
  {
    T pos[vdim + 1];
    for( int i = 0; vdim != i; ++i )
      pos[i] = v[i];

    pos[vdim] = 1.0f;
    *this = Mat( vdim + 1, 1, pos );
    
  }
#endif 
  Mat( uint row, uint col, float const *yo) : m_mat( NULL )
  {
    setSize( row, col );
    uint togo( row * col );
    for( int count = 0; togo != count; ++count )
      m_mat[count] = yo[count];
  }
  
  Mat( Mat const &a ) : m_mat( 0 ) // we need that, set size checks for it before deleting
  {
    size( a.size() );
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      m_mat[c] = a.m_mat[c];
  }
  ~Mat()
  {
    if( m_mat )
      delete [] m_mat;
  }
  /**
   * init gradualy a vector (say, [3, 1] will be .25, .5, .75)
   */
  void initStep();

  Mat &operator=( Mat const &a )
  {
    size( a.size() );
    for( int c = 0, togo = a.m_size[0] * a.m_size[1]; togo != c; ++c )
      m_mat[c] = a.m_mat[c];

  }

  uint m_size[2];

  Mat<T> &f( T f( T const & , T const & ), Mat<T> const &a );
  Mat<T> &operator-=( Mat<T> const & a ) ;
  Mat<T> const operator-( Mat<T> const & a ) const;

  static inline Mat<T> rotX( T const &alpha );
  static inline Mat<T> rotY( T const &alpha );
  static inline Mat<T> rotZ( T const &alpha );

  static inline Mat<T> rot( T const &x, T const &y, T const &z );
  static inline Mat<T> translate( T const &x, T const &y, T const &z );
  /*\
   * @brief transpose
   */
  Mat<T> const t() const;

  /*\
   * @brief Invert matrice.
   */
  Mat<T> const i() const;

  void rand( T const & low, T const & hi );
  
  inline Mat<T> f( T f( T const & ) )
  {
    Mat<T> rv( *this );
    
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      rv.m_mat[c] = f( rv.m_mat[c] );

    return rv;
  }
  inline Mat<T> f( T f( T ) )
  {
    Mat<T> rv( *this );
    
    for( int c = 0, togo = m_size[0] * m_size[1]; togo != c; ++c )
      rv.m_mat[c] = f( rv.m_mat[c] );

    return rv;
  }
  static inline Mat<T> Fixed( uint r, uint c, T f() );
  static inline Mat<T> Ones( uint const *size );
  static inline Mat<T> Zeros( uint r, uint c );
  static inline Mat<T> I( uint r, uint c );


  uint const inline *size() const { return getSize(); };

  inline void size( uint n ) { size( n, 1 ); };
  uint const inline *getSize() const { return m_size; }
  
  Mat<T> const &operator*=( T const & d );

  Mat<T> const operator*( T const & d ) const
  {
    Mat<T> rv( *this );

    rv *= d;
    return rv;
  }

  inline void size( uint const *size )
  {
    setSize( size[0], size[1] );
  }
  inline void size( uint r, uint c )
  {
    setSize( r, c );
  }
  inline void setSize( uint i, uint j )
  {
    //if( ( m_size[0] == i ) 
    //	&& (m_size[1] == j) )
    //  return;

    m_size[0] = i;
    m_size[1] = j;
    if( m_mat )
      delete [] m_mat;

    m_mat = new T[ i * j ];
  };

  T & operator()( uint i, uint j ) const
  {
    return m_mat[ i * m_size[1] + j];
  }


  
  struct Row
  {
 
    T &operator[]( uint i )
    {
      assert( i < size );
     
      return m_[i];
    }
    T operator[]( uint i ) const
    {
     assert( i < size );
     return m_[i];
    }
    T *m_;
    //for debug
    uint size;
    Row( T *p, uint size ) : m_( p ), size(size) {};
  };

  Row operator[]( uint i )
  {
    assert( i < m_size[0] );
    return Row( m_mat + i * m_size[1], m_size[1] );
  }
  Row const operator[]( uint i ) const
  {
    assert( i < m_size[0] );
    return Row( m_mat + i * m_size[1], m_size[1] );

  }

  Mat<T> const operator*( Mat<T> const & a ) const
  {
    Mat<T> rv = Zeros( size()[0], a.size()[1] );

    assert( size()[1] == a.size()[0] );
    for( uint colRes = 0; rv.size()[1] != colRes; ++colRes )
      for( uint rowRes = 0; rv.size()[0] != rowRes; ++rowRes )
	for( uint intCount = 0; a.size()[0] != intCount; ++intCount )
	{
	  float const &t1( (*this)[rowRes][intCount] );
	  Row const &arow( a[intCount] );
	  float const &t2( arow[colRes] );

	  float &store( rv[rowRes][colRes] );
	  
	  store += t1 * t2;
	}
    
    return rv;
  }

  //  private:
  T * m_mat;
};

template <typename T>
Mat<T> const & Mat<T>::operator*=( T const & d )
{
  for( uint i = 0, togo( m_size[0] * m_size[1] ); togo != i; ++i )
    m_mat[i] *= d;

  return *this;
}

template <typename T>
Mat<T> const Mat<T>::t() const
{
  Mat<T> rv;
  rv.size( m_size[1], m_size[0] );

  for( int i = 0; m_size[0] != i; ++i )
    for( int j = 0; m_size[1] != j; ++j )
      rv[j][i] = (*this)[i][j];

  return rv;
}

template <typename T>
Mat<T> const Mat<T>::i() const
{
  Mat<T> const &A( *this );

  unsigned int const n( A.size()[0] );

  Mat<T> temp;
  temp.size( n, 2 * n );

  for( int i = 0; n != i; ++i )
    for( int j = 0; n != j; ++j )
      temp[i][j] = A[i][j];

  for( int i = 0; n != i; ++i )
    for( int j = 0; n != j; ++j )
      temp[i][j + n] = i == j ? 1.0f : .0f;

  std::cout << temp << std::endl;
  GaussianElimination( &temp );
  std::cout << temp << std::endl;
  ReduceEchelon( &temp );
  std::cout << temp << std::endl;

  Mat<T> rv;
  rv.size( A.size() );

  for( int i = 0; n != i; ++i )
    for( int j = 0; n != j; ++j )
      rv[i][j] = temp[i][j + n];



  return rv;
}

template <typename T>
Mat<T> &Mat<T>::f( T f( T const &, T const & ), Mat<T> const &a )
{
  assert( a.m_size[0] == m_size[0] );
  assert( a.m_size[1] == m_size[1] );

  int const togo( m_size[0] * m_size[1] );
  for( int c = 0; togo != c; ++c )
    m_mat[c] = f( m_mat[c], a.m_mat[c] );

  return *this;
}

template <typename T>
T minus( T const &, T const & );

template <typename T>
Mat<T> &Mat<T>::operator-=( Mat<T> const & a ) 
{
  return this->f( minus, a );
}

template <typename T>
Mat<T> const Mat<T>::operator-( Mat<T> const & a ) const
{
  Mat<T> rv( *this );

  return rv -= a;
}


#if 0
template <typename T>
Mat<T> const Mat<T>::operator*( Mat<T> const & a ) const
{
  Mat<T> rv = Zeros( size()[0], a.size()[1] );

  assert( size()[1] == a.size()[0] );
  for( uint colRes = 0; rv.size()[1] != colRes; ++colRes )
    for( uint rowRes = 0; rv.size()[0] != rowRes; ++rowRes )
      for( uint intCount = 0; a.size()[0] != intCount; ++intCount )
	rv[rowRes][colRes] += (*this)[rowRes][intCount] * a[intCount][colRes];

  return rv;
}
#endif

template <typename T>
T One();

template <typename T>
T Zero();


template <typename T>
Mat<T> Mat<T>::Fixed( uint r, uint c, T f() )
{
  Mat<T> rv;
  rv.size( r, c );
  
  for( uint i = 0, togo( r * c ); i != togo; ++i )
    rv.m_mat[i] = f();
  return rv;
}

template <typename T>
Mat<T> Mat<T>::Ones( uint const *size )
{
  return Fixed( size[0], size[1], One );
}

template <typename T>
Mat<T> Mat<T>::Zeros( uint r, uint c )
{
  Mat<T> rv( Fixed( r, c, Zero) );
  return rv;
}

template <typename T>
Mat<T> Mat<T>::I( uint r, uint c )
{
  Mat<T> rv;
  rv.size( r, c );
  
  for( int i = 0; r != i; ++i )
    for( int j = 0; c != j; ++j )
      rv[i][j] = ( i == j ? One<T>(): Zero<T>() );
  /*
  ** maxima 
  ** matrix( [5,0,10],  [0,10,0],  [10,0,34]);
  */
  return rv;
}

//template <>
//Mat<T>

template <typename T>
Mat<T> Mat<T>::rotX( T const & a )
{

  float const cos( cosf( a ) );
  float const sin( sinf( a ) );

  T m9[4][4] = {
    { 1.0f, .0f, .0f, 0 },
    { 0, cos, sin, 0 },
    { 0, - sin, cos, 0 },
    { 0, 0, 0, 1 }
  };

  return Mat( 4, 4, (float const *)m9 );
}

template <typename T>
Mat<T> Mat<T>::rotY( T const & a )
{

  float const cos( cosf( a ) );
  float const sin( sinf( a ) );

  T m9[][4] = {
    { cos, 0, sin, 0 },
    { .0f, 1.0f, .0f, 0 },
    { - sin, .0f, cos, 0 },
    { 0, 0, 0, 1 },
  };

  return Mat( 4, 4, (float const *)m9 );
}

template <typename T>
Mat<T> Mat<T>::rotZ( T const & a )
{

  float const cos( cosf( a ) );
  float const sin( sinf( a ) );

  T m9[4][4] = {
    { cos, - sin, 0, 0 },
    { sin, cos, 0, 0 },
    { .0f, .0f, 1.0f, 0 },
    { .0f, .0f, .0f, 1.0f },
  };

  return Mat( 4, 4, (float const *)m9 );
}

template <typename T>
Mat<T> Mat<T>::rot( T const & x, T const & y, T const & z )
{
  return rotX( x ) * rotY( y ) * rotZ ( z );
}

template <typename T>
Mat<T> Mat<T>::translate( T const & x, T const & y, T const & z )
{
  float p[4][4] = {
    { 1, 0, 0, x },
    { 0, 1, 0, y },
    { 0, 0, 1, z },
    { 0, 0, 0, 1 },
  };

  Mat<T> rv( 4, 4, (float const *)p );

  return rv;
}


template <>
inline float Zero<float>()
{
  return 0.0f;
}

template <typename T>
inline void swap( T &a, T &b )
{
  T temp( a );
  a = b;
  b = temp;  
}

template <typename T> 
std::basic_ostream<char> &
operator<<( std::basic_ostream<char> &o, Mat<T> const & m )
{
  for( int i = 0; m.m_size[0] != i; ++i )
  {
    o << ( i ? "   " : "[  ");

    for( int j = 0; m.m_size[1] != j; ++j )
    {
      o << m(i, j);
      if( ( m.m_size[1] - 1 ) != j )
	o << ", ";
    }
	
    o << ( ( ( m.m_size[0] - 1 ) == i ) ? "  ]" : ";" );
    o << std::endl;
  }

  return o;
}

/*
** @germ
** from wikipedia
*/ 
template <typename t>
void GaussianElimination( Mat<t> *mat )
{
  Mat<t> &A( *mat ); // prototype as pointer because value is R/W
  int const m( A.size()[0] );
  int const n( A.size()[1] );

  int i( 0 );
  int j( 0 );

  while( ( m != i )
	 && ( n != j ) )
  {
    int maxi( i );
    for( int k = i + 1; m != k; ++k )
      if( abs( A[k][j] > abs( A[maxi][j] ) ) )
	maxi = k;
    
    static const float e( .001f );
    if( abs( A[maxi][j] ) > e )
    {
      // swap rows i and maxi, but do not change the value of i
      for( int k = 0; n != k; ++k )
	swap( A[i][k], A[maxi][k] );
      // Now A[i,j] will contain the old value of A[maxi,j].


      // divide each entry in row i by A[i,j]
      t const Aij( A[i][j] );
      for( int k = 0; n != k; ++k )
	A[i][k] /= Aij;
      //Now A[i,j] will have the value 1.

      for( int u = i + 1; m != u; ++u )
      {
	t const Auj( A[u][j] );
	for( int k = 0; n != k; ++k )
	  A[u][k] -= Auj * A[i][k];
	// Now A[u,j] will be 0, since A[u,j] - A[i,j] * A[u,j] = A[u,j] - 1 * A[u,j] = 0.
      }
      ++i;
    }
    ++j;
  }
}

/*
** fixme: factorise with SolveSystem ? @germ
*/
template <typename t>
void ReduceEchelon( Mat<t> *mat )
{
  Mat<t> &A( *mat );
  unsigned int const n( A.size()[0] );
  unsigned int const m( A.size()[1] );

  // init z
  t const Ann( A[n - 1][n - 1] );
  for( int k = 0; m != k; ++k )
    A[n - 1][k] /= Ann;

  for( int i = n - 2; i >= 0; --i )
  {
    for( int j = n - 1; j != i; --j )
    {
      t const Aij( A[i][j] );
      for( int k = 0; m != k; ++k )
	A[i][k] -= Aij * A[j][k];
    }
  }
}

template <typename t>
void SolveLinearSystem( Mat<t> *mat )
{
 
  Mat<t> &A( *mat );

  unsigned int const m( A.size()[0] );
  unsigned int const n( A.size()[1] );

  GaussianElimination( &A );
  ReduceEchelon( &A );  
}

template <typename t, int dim>
inline t staticPow( t const &x );

template <>
inline float staticPow<float,0>( float const & )
{
  return .0f;
}

template <typename t, int dim>
inline t staticPow( t const &x )
{
  return x * staticPow<t,dim - 1>( x );
}

/*\
 * @brief  pass it a matrix with the tuple (x,y) as rows, it gives you the 
 * a0 .. an coefficient to connect your points using the polynom
 * y = a0 + a1 * x + .. an * xn
 */
template <typename t>
void SolveNormalSystem( Mat<t> const &xyTuple, Mat<t> *as )
{
  int const nPoints( xyTuple.size()[0] );
  int const degree( nPoints - 1);
  assert( degree >= 2 && "why do you want to solve with a polynom in x ?");

  Mat<t> j, ys;

  j.size( nPoints, degree + 1 ); // ( i know it s the same : but rows are points and column are degrees )
  for( int degreI = 0; degree >= degreI; ++degreI )
    for( int pointI = 0; nPoints != pointI; ++pointI )
      j[pointI][degreI] = pow( xyTuple[pointI][0], degreI ); // fixme: break the template
  std::cout << j << std::endl;
  ys.size( nPoints, 1 );
  for( int pointI = 0; nPoints != pointI; ++pointI )
    ys[pointI][0] = xyTuple[pointI][1];
  std::cout << ys << std::endl;

  Mat<t> const jt( j.t() );
  std::cout << jt << std::endl;

  Mat<t> const jtj( jt * j );
  std::cout << jtj << std::endl;

  Mat<t> const jtj__1( jtj.i() );
  std::cout << jtj__1 << std::endl;

  Mat<t> const E1( ( j * jtj__1 ).t() );
  std::cout << E1 << std::endl;
  *as = E1 * ys ;
}

template <typename t>
t lerp( t const &x0, t const &x1, t const &y0, t const &y1, t const &x )
{
  return ( x - x0 ) * ( ( y1 - y0 ) / ( x1 - x0 ) ) + y0; 
}

template <typename t, int _degree>
class SmoothPath
{
  /*
  ** we are allways looking for two polynom because we are in 3
  ** dimension (I use the vector class defined earlier) 
  */

  Mat<t> *m_a;
  Vector<t> *m_points;
  unsigned int m_nPoints;

  struct LocalAxis
  {
    Vector<t> m_localAxisOrigin;
    Vector<t> m_localAxisUnit[3];
  } *m_localAxis;

public:
  SmoothPath() : m_a( NULL ), m_points( NULL ), m_localAxis( NULL )
  {
  };

  ~SmoothPath()
  {
    if( m_a )
      delete [] m_a;
    if( m_points )
      delete [] m_points;
    if( m_localAxis )
      delete [] m_localAxis;
  }


void
  Solve( Vector<t> const *points, unsigned int const nPoints )
  {
    if( m_a )
      delete [] m_a;
    if( m_points )
      delete [] m_points;
    if( m_localAxis )
      delete [] m_localAxis;

    m_nPoints = nPoints;
    m_points = new Vector<t>[nPoints];

    memcpy( m_points, points, nPoints * sizeof( Vector<t> ) ); // safe for copy

    assert( _degree < nPoints );
    static unsigned int const pointsPerCurve( _degree + 1 );

    Mat<t> xys;
    xys.size( pointsPerCurve, 2 );
    /*
    ** say we have three point and look for a quadric, we have one curve
    ** if we will have one more curve per extra points.
    */
    unsigned int const nCurve( nPoints - _degree ); 


    m_a = new Mat<t>[nCurve * 2];
    m_localAxis = new LocalAxis[nCurve];

    /*
    ** so, now we solve, curve by curve
    */
    for( unsigned int curveI = 0; nCurve != curveI; ++curveI )
    {
      Vector<t> const localAxis[2] = 
	{ 
	  points[curveI],
	  points[curveI + _degree]
	};

      Vector<t> const localAxisUnitX( ( localAxis[1] - localAxis[0] ).unit() );

      /*
      ** compute secondary axis based on second point
      */
      Vector<t> const &A( points[curveI] );
      Vector<t> const &B( points[curveI + 1] );
      Vector<t> const D( points[curveI + _degree] );
      // fixme: I think i went to strong on Y
      Vector<t> const C( ( D - A ).unit() * ( ( B - A ) * ( D - A ).unit() ) + A );

      Vector<t> const localAxisUnitY( ( B - C ).unit() );
      Vector<t> const localAxisUnitZ( localAxisUnitX ^ localAxisUnitY );
      //t const lengthZ( localAxisUnitZ.length() );
      //assert( 1.0f == lengthZ );


      LocalAxis curveAxis = { localAxis[0], { localAxisUnitX, localAxisUnitY, localAxisUnitZ } };
      m_localAxis[curveI] = curveAxis;
      Vector<t> localPoints[_degree + 1];
      for( unsigned int pointI = 0; pointsPerCurve != pointI; ++pointI )
	localPoints[pointI] = Vector<t>( ( points[curveI + pointI] - localAxis[0] ) * localAxisUnitX,
					 ( points[curveI + pointI] - localAxis[0] ) * localAxisUnitY,
					 ( points[curveI + pointI] - localAxis[0] ) * localAxisUnitZ );

      for( unsigned short polyI = 0; 2 != polyI; ++polyI )
      {
	static unsigned short const s_localYIndexPerPoly[2] = { 1, 2 };
	for( unsigned int pointI = 0; pointsPerCurve != pointI; ++pointI )
	{
	  assert( localPoints[pointI][0] >= .0f && "degenerated path" );

	  xys[pointI][0] = localPoints[pointI][0];
	  xys[pointI][1] = localPoints[pointI][s_localYIndexPerPoly[polyI]];
	
	}
	std::cout << "xys: " << std::endl << xys << std::endl;
	SolveNormalSystem( xys, &m_a[curveI * 2 + polyI] );
	std::cout << "a: " << std::endl << m_a[curveI * 2 + polyI] << std::endl;
      }
    }
  };

  Vector<t> 
  GetPositionT( float time )
  {
    t totalDist( .0f );
    assert( m_nPoints > 0 );
    for( int pointI = 1; m_nPoints != pointI; ++pointI )
      totalDist += m_points[pointI - 1].dist( m_points[pointI] );

    int currPointI( 0 );
    t currDist( .0f );
    t previousDist( .0f );
    for( ; ( ( m_nPoints - 1) != currPointI ) && ( ( currDist / totalDist ) < time ) ;
	 ++currPointI )
    {
      previousDist = currDist;
      currDist += m_points[currPointI].dist( m_points[currPointI + 1] );
    }
    --currPointI; // fix that,, nicer
    glColor3f( 1, 1, 1);
    glBegin( GL_LINES );
    {
      // glVertex3fv( &m_points[currPointI][0]);
      //glVertex3fv( &m_points[currPointI + 1][0]);
    } glEnd();

    t const localDist( time * totalDist - previousDist );
    Vector<t> const protoPoint( ( m_points[currPointI + 1] - m_points[currPointI] ).unit() * localDist + m_points[currPointI] );
    Vector<t> sumGlobalCoordinate( .0f, .0f, .0f );
    t totalCoef( .0f );
 
    for( int degreI = 0; _degree != degreI; ++degreI )
    {
      int const curveIndex( currPointI - ( _degree - 1 ) + degreI );
      if( 0 > curveIndex )
	continue;
      if( ( curveIndex + _degree ) >= m_nPoints )
	continue;

      Vector<t> const curveAxis( ( m_points[curveIndex + _degree] - m_points[curveIndex] ).unit() );
      t const x__( ( protoPoint - m_localAxis[curveIndex].m_localAxisOrigin ) * curveAxis );
      t const x( x__ );

      t const xmiddeg( ( m_points[curveIndex + _degree / 2 ] - m_localAxis[curveIndex].m_localAxisOrigin ) * curveAxis );
      t const xf( ( m_points[curveIndex + _degree ] - m_localAxis[curveIndex].m_localAxisOrigin ) * curveAxis );

      t poly[2] = { .0f, .0f };

      for( int polyI = 0; 2 != polyI; ++polyI )
      {
	Mat<t> const a( m_a[curveIndex * 2 + polyI] );
	for( int degreI = 0; _degree >= degreI; ++degreI )
	  poly[polyI] += pow( x, degreI ) * a[degreI][0];
      }


      // polyAllDim  contains the position for local t in local coordinate
      glColor3f( 0, 0, 1);
      glBegin( GL_LINES );
      {
	//glVertex3fv( &(m_localAxis[curveIndex].m_localAxisOrigin)[0] );
	//glVertex3fv( &(m_localAxis[curveIndex].m_localAxisOrigin+ m_localAxis[curveIndex].m_localAxisUnit[0])[0] );
	//glVertex3fv( &(m_localAxis[curveIndex].m_localAxisOrigin)[0] );
	//glVertex3fv( &(m_localAxis[curveIndex].m_localAxisOrigin+ m_localAxis[curveIndex].m_localAxisUnit[1])[0] );
	
      } glEnd();
      t const coef( x > xmiddeg ? lerp( xmiddeg, xf, 1.0f, .0f, x ) : lerp( .0f, xmiddeg, .0f, 1.0f, x ) );
      Vector<t> const globalCoordinate( ( m_localAxis[curveIndex].m_localAxisOrigin 
					  + m_localAxis[curveIndex].m_localAxisUnit[0] * x
					  + m_localAxis[curveIndex].m_localAxisUnit[1] * poly[0]
					  + m_localAxis[curveIndex].m_localAxisUnit[2] * poly[1] )
					* coef );
      totalCoef += coef;
 
      sumGlobalCoordinate += globalCoordinate;
    }
    return sumGlobalCoordinate / totalCoef;
  }
};


typedef Vector<float> Vectorf;
typedef Mat<float> Matf;



int 
main( int argc, char *argv[] )
{

  //  std::cout << a << std::endl;

  int I = 0;

  static const Vectorf points[] = 
    {
      //      Vectorf( 0, 2.6, 0 ),
      //     Vectorf( 0, 2.6, 0 ),
      Vectorf( 1, 4, 0 ),
      Vectorf( 2, 3, 0 ),
      Vectorf( 1.7, 2, 0 ),
      Vectorf( 3, 1, 0 ),
      Vectorf( 2.5, 0, 0 ),
      Vectorf( 1.5, -1, 0 ),
      Vectorf( -1.5, -3, 0 ),
      Vectorf( -2.5, -2.5, 0 ),
     


    };
  static const int nPoints( sizeof points / sizeof *points );
  Matf a;
  Matf xys( 4, 3, &points[0][0]); // fixme, not great for the third column

  SmoothPath<float,3> path;

  path.Solve( points, nPoints );

  SolveNormalSystem( xys, &a );
  std::cout << "a:" << a << std::endl;


  {
    /*
    ** wikipedia test gauss
    */
    static const float _mat[][4] = 
      {
	{  2,  1, -1,   8 },
	{ -3, -1,  2, -11 },
	{ -2,  1,  2,  -3 },
      };

    Matf mat( 3, 4, (float*)_mat );

    //std::cout << mat << std::endl;
    SolveLinearSystem( &mat );
    //    std::cout << mat << std::endl;
    int I = 0; //bp here
  }
  {
    /*
    ** wiki line smoothing
    */ 
    static const float _jt[3][5] =
      {
	{ 1, 1, 1, 1, 1 },
	{ - 2, - 1, 0, 1, 2 },
	{ 4, 1, 0, 1, 4 },
      };
    Matf const jt( 3, 5, (float*)_jt );
    Matf const j( jt.t() );
    Matf const jtj( jt * j );

    //std::cout << "jtj:" << std::endl;
    //std::cout << jtj << std::endl;

    Matf const jtj__1( jtj.i() );

    //std::cout << "jtj^-1:" << std::endl;
    //std::cout << jtj__1 << std::endl;

    Matf const E1( ( j * jtj__1 ).t() );

    //std::cout << "E1" << std::endl;
    //std::cout << E1 << std::endl;

    

    //std::cout << "done" << std::endl;


  }
  
  SDL_Event event;
  SDL_Surface *screen;
  bool done( false );
  SDL_Init( SDL_INIT_VIDEO );
 
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  screen = SDL_SetVideoMode(640, 480, 8, SDL_OPENGL );

  glScalef( .2, .2, .2 );

  float t( .0f );

  while( !done )
  {
    t += .001f;
    if( t > 5.0f )
      t -= 5.0f;
    //    t = 1.75;
    glClear( GL_COLOR_BUFFER_BIT );
    Vectorf const interp2( path.GetPositionT( t / 5.0f ) );
    glColor3f( 0, 1, 0 );
    glBegin( GL_POINTS );
    {
      for( int i = 0; nPoints != i; ++i )
	glVertex3fv( &points[i][0] );

      //      float const y( a[0][0] + t * a[1][0] + sqr( t ) * a[2][0] + powf( t, 3 ) * a[3][0] );
      float y( 0 );
      for( int ai = 0; a.size()[0] != ai; ++ai )
	y += pow(t, ai) * a[ai][0];
      
      Vectorf interp( t, y, 0 );

      //glVertex3fv( &interp[0] );
      glColor3f( 1, 1, 1 );
      glVertex3fv( &interp2[0] );
    } glEnd();
    SDL_GL_SwapBuffers( );

    while( SDL_PollEvent( &event ) )
    {
      static bool left( false );
      static bool right( false );
      static bool middle( false );


      switch( event.type )
      {
	
      case SDL_KEYDOWN :
	break;
      case SDL_QUIT:
	done = true;
	break;
      default:
	break;
      }
    }
  }

  SDL_Quit();

  return 0;
}
