//(c) Germain Le Chapelain 2007

template <typename T>
class Vertex
{
  T m_pos[3];
public:
  Vertex(){};

  Vertex( T const &x, T const &y, T const &z )
  {
    m_pos[0] = x;
    m_pos[1] = y;
    m_pos[2] = z;
  }

  /*\
   * @brief dot product
   */

  inline T &operator*=( Vertex const &a ) const
  {
    for( int i = 0; 3 != i; ++i )
      *this[i] *= a[i];

    return *this;
  }

  inline T const dot( Vertex const &a ) const
  {
    T returnValue( .0f ); // fixme;template
    for( int i = 0; 3 != i; ++i )
      returnValue += (*this)[i] * a[i];

    return returnValue;
  }

  inline T const operator*( Vertex const &a ) const
  {
    return this->dot( a );
  }

  /*\
   * @brief cross product
   */
  inline Vertex cross( Vertex const &v ) const
  {
    return Vertex( (*this)[1] * v[2] - (*this)[2] * v[1], 
		   (*this)[2] * v[0] - (*this)[0] * v[2],
		   (*this)[0] * v[1] - (*this)[1] * v[0] );
  }
  inline Vertex operator^( Vertex const & v ) const
  {
    return this->cross( v );
  }
  inline Vertex &operator+=( Vertex const & v )
  {
    for( int i = 0; 3 != i; ++i )
      m_pos[i] += v[i];
    return *this;
  }

  inline Vertex operator+( Vertex const & v ) const
  {
    Vertex rv( *this );
    return rv += v;
  }

  inline Vertex &operator*=( T const & a )
  {
    for( int i = 0; 3 != i; ++i)
      m_pos[i] *= a;
    return *this;
  }

  inline Vertex &operator-=( Vertex const &a )
  {
    return (*this) += a * -1.0f;
  }

  inline Vertex operator-( Vertex const &a ) const
  {
    return *this + ( a * -1.0f );
  }

  inline Vertex operator*( T const & a ) const
  {
    Vertex rv( *this );
    return rv *= a;
  }

  inline Vertex operator/( T const & a ) const
  {
    return *this * ( 1.0f / a );
  }
  
  inline T dist( Vertex const &a ) const
  {
    return sqrt( dist2( a ) );
  }

  inline T length() const
  {
    return dist( Vertex( 0, 0, 0) );
  }

  inline Vertex unit() const
  {
    return *this / length();
  }

  inline T dist2( Vertex const &a ) const
  {
    T rv( .0f );

    for( int i = 0; 3 != i; ++i )
      rv += sqr( m_pos[i] - a.m_pos[i] );

    return rv;
  }

  inline T const &operator[]( int indice ) const { return m_pos[indice]; };
  inline T &operator[]( int indice ) { return m_pos[indice]; };
};
